// Soubor DEDIC4.CPP
// Jednoduchy priklad pouziti virtualnich metod
// Program napodobuje graficky editor: Obsahuje hierarchii grafickych objektu
//
// Spolecnym predkem je trida GO (tj. graficky objekt), ktera 
// implementuje spolecne vlastnosti - barvu (jako cislo typu unsigned long)
// pristupovou metodu Barva(), ktera tuto barvy vraci, a metody
// Nakresli() a Smaz()
// Tyto metody jsou virtualni (kazdy graficky objekt se kresli - a take maze - jinak
// Navic jsou ciste virtualni (nemaji implementaci), nebot obecny graficky objekt nelze nakreslit
//
// Od tridy GO jsou odvozeny instancni tridy bod a kruh
// tyto tridy maji navic datove slozky (souradnice bodu, stred a polomer kruhu)
// a implementuji metody Nakresli() a Smaz()
// specifikaci virtual neni v jejich deklaraci treba opakovat
//
// Abychom se vyhnuli technickym detailum kresleni (to C++ neresi, to zavisi na prostredi, 
// pro ktere je program urcen), implementujeme tyto metody tak, ze proste vypiseme na obrazovku
// "kreslim bod..." , "kreslim kruh ..." 
//
// Program pracuje s dynamickymi instancemi trid bod a kruh. Jejich adresy uklada do
// pole ukazatelu na graficke objekty (tj. na typ GO)
// Polymorfizmus umoznuje nakreslit nebo smazat vsechny objekty v jednom cyklu


#include <iostream.h>

class GO {
public:
	GO(unsigned long _barva):barva(_barva){}
	virtual void Nakresli() = 0;
	virtual void Smaz() = 0;
	virtual ~GO(){ cout << "rusim obecny GO" << endl;}
	unsigned long Barva() {return barva;}
private:
	unsigned long barva;
};

class bod: public GO {
	double x, y;		// Souradnice
public:
	bod(double _x, double _y, unsigned long _barva)
		: GO(_barva), x(_x), y(_y) {}
	~bod(){ Smaz(); }						// Specifikaci virtual neni treba opakovat
	void Nakresli()
	{
		cout << "kreslim bod" << "(" << x << ", " << y << ")"<< " barva: " << Barva()  << endl;
	}
	void Smaz()
	{
		cout << "mazu bod" << "(" << x << ", " << y << ")"<< " barva: " << Barva()  << endl;
	}
};

class kruh: public GO {
	double x, y, r;	// Souradnice stredu a polomer
public:
	kruh(double _x, double _y, double _r, unsigned long _barva)
		: GO(_barva), x(_x), y(_y), r(_r) {}
	~kruh(){ Smaz(); }
	void Nakresli()
	{
		cout << "kreslim kruh" << "(" << x << ", " << y << ", " << r << ")"<< " barva: " << Barva()  << endl;
	}
	void Smaz()
	{
		cout << "mazu kruh" << "(" << x << ", " << y << ", " << r << ")"<< " barva: " << Barva()  << endl;
	}
};


void main()
{
	GO *go[10];
	go[0] = new bod(1,1,0);
	go[1] = new kruh(2,5,8,9);
    int i;
	for(i = 0; i < 2; i++) go[i] ->Nakresli();
	for(i = 0; i < 2; i++) delete go[i];
}
