// Soubor T_SEZN1.CPP
// Sablona objektoveho typu: jednosmerny seznam se zarazkou
// Parametr T urcuje typ hodnot ukladanych do seznamu 

// Nejprve deklarujeme sablony prvku seznamu, typ prvek. 
// Ve tride prvek<T> ovsem chceme deklarovat tridu seznam<T>
// jako pritele, a protoze specifikace friend 
// nenahrazuje deklaraci typu (ani predbeznou), musime jako prvni uvest 
// predbeznou deklaraci sablony 
//
// Trida seznam obsahuje konstruktor, destruktor a metody
// - VlozNaKonec: Vlozi zadanou hodnotu typu T na konec seznamu
// - Vyprazdni - odstrani ze seznamu vsechny prvky krome zarazky
// - VyjmiPrvni - vyjme ze seznamu prvni prvek a vrati hodnotu v nem ulozenou
// - Prazdny - vrati true, je-li seznam prazdny, tj. obsahuje-li pouze zarazku
//
// Dale obsahuje 
// staticky atribut pocet, ktery v kazdem okamziku urcuje pocet existujicich 
// instanci tridy seznam,
// a statickou metodu Pocet(), ktera vraci hodnotu atributu pocet.

#include <string>
#include <iostream>
#include <cstdlib>

using namespace std;

//template<class T> class seznam;	// Predbezna deklarace sablony

template<class T>
class prvek {						// Prvek seznamu
	T data;
	prvek* nasl;
	//friend class seznam<T>;			// Trida seznam smi porusovat pristupova prava
public:
	prvek(T t): data(t), nasl(0) {}
	T& hodnota() { return data; }   // Pristupove metody jsou referencni - lze je pouzit i
	prvek*& dalsi() { return nasl; }// k prirazovani hodnot (zapsat na prave strane prirazeni)
};

template<class T>						// Sablona tridy seznam
class seznam {
public:
	seznam();
	~seznam();
	void Vyprazdni();
	void VlozNaKonec(T co);
	T VyjmiPrvni();
	bool Prazdny() { return hlava == zarazka; }
	static int Pocet() { return pocet; }
private:
	prvek<T> *hlava, *zarazka;
	static int pocet;					// Aktualni pocet seznamu (instanci tridy seznam)
	void Chyba(string s){				// Primitivni osetreni chyby
		cout << s << endl;
		exit(0);
	}
};


// Sablona statickeho atributu 
template<class T> int seznam<T>::pocet = 0;

// Sablona konstruktoru
template<class T>
seznam<T>::seznam(): hlava(0), zarazka(0) {
	hlava = new prvek<T>(0);
	if(hlava) 
	{
		zarazka = hlava;
		pocet++;
	}
	else 
	{ 
		hlava = 0;
		Chyba("Nepodarila se alokace v konstruktoru");
	}
};

template<class T>
seznam<T>::~seznam() {
	Vyprazdni();
	delete zarazka;
	pocet--;
};

template<class T>
void seznam<T>::Vyprazdni(){
	prvek<T>* p = hlava;
	while(hlava != zarazka)
	{
		hlava = hlava->dalsi();
		delete p;
	}
};

template<class T>
void seznam<T>::VlozNaKonec(T co){
	zarazka -> hodnota() = co;
	zarazka -> dalsi() = new prvek<T>(0);
	if(!(zarazka -> dalsi())) Chyba("Nepodarila se alokace ve vloz()");
	zarazka = zarazka -> dalsi();
};

template<class T>
T seznam<T>::VyjmiPrvni()
{
	if(hlava == zarazka) Chyba("Vyber z prazdneho seznamu");
	prvek<T>* p = hlava;
	hlava = hlava -> dalsi();
	T t = p -> hodnota();
	delete p;
	return t;
};


void main()
{
	seznam<int> S, T, U;
	for(int i = 0; i < 10; i++)
		S.VlozNaKonec(i+5);
	while(!S.Prazdny())
		cout << S.VyjmiPrvni() << "  ";
	cout << endl;
	cout << "Pocet seznamu typu int: " << seznam<int>::Pocet() << endl;
	cout << "Pocet seznamu typu double: " << seznam<double>::Pocet() << endl;
}