// Soubor T_SEZNAM.CPP
// Sablona objektoveho typu: jednosmerny seznam se zarazkou
// Parametr T urcuje typ hodnot ukladanych do seznamu 
//
// Trida seznam obsahuje konstruktor, destruktor a metody
// - VlozNaKonec: Vlozi zadanou hodnotu typu T na konec seznamu
// - Vyprazdni - odstrani ze seznamu vsechny prvky krome zarazky
// - VyjmiPrvni - vyjme ze seznamu prvni prvek a vrati hodnotu v nem ulozenou
//
// Dale obsahuje definici typu prvek (prvek seznamu, ktery muze obsahovat data typu T)
// a staticky atribut pocet, ktery v kazdem okamziku urcuje pocet existujicich 
// instanci tridy seznam

#include <string>
#include <iostream>
#include <cstdlib>

using namespace std;

template<class T>						// Sablona tridy seznam
class seznam {
public:
	class prvek {						// Vnoreny typ - prvek seznamu
		T data;
		prvek* nasl;
		friend class seznam<T>;			// Trida seznam smi porusovat pristupova prava
	public:
		prvek(T t): data(t), nasl(0) {}
		T& hodnota() { return data; }   // Pristupove metody jsou referencni - lze je pouzit i
		prvek*& dalsi() { return nasl; }// k prirazovani hodnot
	};

	seznam();
	~seznam();
	void vyprazdni();
	void VlozNaKonec(T co);
	T VyjmiPrvni();
private:
	prvek *hlava, *zarazka;
	static int pocet;					// Aktualni pocet seznamu (instanci tridy seznam)
	void Chyba(string s){				// Primitivni osetreni chyby
		cout << s << endl;
		exit(0);
	}
};

// Sablona statickeho atributu (presneji: specifikace implicitni hodnoty)
template<class T> int seznam<T>::pocet = 0;

// Sablona konstruktoru
template<class T>
seznam<T>::seznam(): hlava(0), zarazka(0) {
	hlava = new prvek(0);
	if(hlava) 
	{
		zarazka = hlava;
		pocet++;
	}
	else 
	{ 
		hlava = 0;
		Chyba("Nepodarila se alokace v konstruktoru");
	}
};

template<class T>
seznam<T>::~seznam() {
	vyprazdni();
	delete zarazka;
	pocet--;
};

template<class T>
void seznam<T>::vyprazdni(){
	prvek* p = hlava;
	while(hlava != zarazka)
	{
		hlava = hlava->dalsi();
		delete p;
	}
};

template<class T>
void seznam<T>::VlozNaKonec(T co){
	zarazka -> hodnota() = co;
	zarazka -> dalsi() = new prvek(0);
	if(!(zarazka -> dalsi())) Chyba("Nepodarila se alokace ve vloz()");
	zarazka = zarazka -> dalsi();
};

template<class T>
T seznam<T>::VyjmiPrvni()
{
	if(hlava == zarazka) Chyba("Vyber z prazdneho seznamu");
	prvek* p = hlava;
	hlava = hlava -> dalsi();
	T t = p -> hodnota();
	delete p;
	return t;
};

void main()
{
	seznam<int> S;
	S.VlozNaKonec(10);
}