// Soubor APTR.CPP
// Priklad pouziti automatickych ukazatelu v souvislosti s vyjimkami 
// 
// Trida X ma konstruktor, ktery pri parametru s hodnotou 11 
// vyvola vyjimku
//
// Pozor, pro kontrolni vystup v operatorech new a delete nepouzivat 
// objektove datove proudy - hrozi podivne chyby


#include <stdexcept>
#include <memory>
#include <string>
#include <cstdlib>
#include <cstdio>

using namespace std;

class X {
	int i;
public:
	X(int ii = 0): i(ii){
		printf("Konstruktor X(%d)\n", ii);
		if (ii == 11) throw logic_error("nezadouci hodnota i");
	}
	~X(){ printf("Destruktor X(%d)\n", i); }
};

void main()
{
	try {
		auto_ptr<X> u1(new X(1));
		auto_ptr<X> u2(new X(11));
	}
	catch(exception& e)
	{
		printf("Vyjimka: %s\n", e.what());
	}
}

/* Vystup tohoto programu bude

Konstruktor X(1)
Konstruktor X(11)
Destruktor X(1)
Vyjimka: nezadouci hodnota i

Instance *u1 se alokovala v poradku a vytvoril se pro ni 
automaticky ukazatel, ktery pri vyjimce pro tuto instanci 
zavola operator delete (takze se zavola destruktor a pak se uvolni pamet).

Pro instanci *u2 se vyhradila pamet a pak v konstruktoru vyjimka.
destruktor se tedy nezavola (instance neni plne zkonstruovana),
automaticky ukazatel se nevytvori, ale program automaticky zavola 
operator delete.
*/