// Soubor NEWEX.CPP
// Priklad operatoru new a delete se stejnymi dodatecnymi parametry
// JEN NEJNOVEJSI PREKLADACE
// 
// Trida X ma konstruktor, ktery pri parametru s hodnotou 11 
// vyvola vyjimku
//
// Operator new ma dodatecny parametr typu unsigned char, ktery 
// se pouzije k inicializaci pameti instance.
// Operator delete ma tyz dodatecny parametr.
// Dojde-li pri alokaci dynamicke instance k vyjimce, zavola se pro 
// uvolneni pameti alokovane pro nezkonstruovanou instanci automaticky 
// operator delete s dodatecnymi parametry stejnymi jako byly dodatecne 
// parametry operatoru new.
//
// Pozor, pro kontrolni vystup v operatorech new a delete nepouzivat 
// objektove datove proudy - hrozi podivne chyby

#include <stdexcept>
#include <string>
#include <cstdlib>
#include <cstdio>

using namespace std;

class X {
	int i;
public:
	X(int ii = 0): i(ii){
		printf("Konstruktor X(%d)\n", ii);
		if (ii == 11) throw logic_error("nezadouci hodnota i");
	}
	~X(){ printf("Destruktor X(%d)\n", i); }
};

void *operator new(size_t s, unsigned char co)
{
	void *p = malloc(s);
	memset(p, co, s);
	printf("Operator new\n");
	return p;
}

void operator delete(void* p, unsigned char co)
{
	printf("Operator delete\n");
	free(p);
}

void main()
{
	try {
		X* u1 = new(0xff) X(1);
		delete u1;
		X* u2 = new(0xff) X(11);
	}
	catch(exception& e)
	{
		printf("Vyjimka: %s\n", e.what());
	}
}

/* Vystup dohoto programu bude
Operator new
Konstruktor X(1)
Destruktor X(1)
Operator new
Konstruktor X(11)
Operator delete
Vyjimka: nezadouci hodnota i


Vsimnete si, ze se neuvolnila pamet pro prvni instanci,
ktera se v poradku zkonstruovala, ale jen pro druhou instanci,
pro kterou se sice alokovala pamet, ale v konstruktoru
vznikla vyjimka.
Vsimnete si take, ze prikazem
delete u1 
se nezavolal nas operator delete, ale standardni verze
tohoto operatoru.
*/