// Soubor RETHROW.CPP
// Priklad castecneho osetreni vyjimky
// 
//Funkce Uloz(S, Kolik, Soubor) ma za ukol vyjmout Kolik prvku seznamu  
// S a ulozit je do souboru Soubor.
// Pouzijeme neobjektove nastroje ze <stdio.h>.
//
// Jednotlive prvky ze seznamu vybirame pomoci metody S.VyjmiPrvni, ktera
// muze vyvolat vyjimku chyba_seznamu. Kdyby k takove vyjimce ale doslo, zustal
// by soubor Soubor otevreny, a to nechceme. Proto vyjimku
// zachytime, v handleru soubor uzavreme a pak vyjimku posleme dal.
//
// Elegantnejsi reseni zalozene na objektovych datovych proudech, ktere
// nepotrebuje zachycovat vyjimku a pak ji posilat dal, ukazuje funkce Uloz2():
// Vytvorime proud (objekt typu ofstream z hlavickoveho souboru <fstream>), 
// vybirame prvky jeden po druhem a ukladame je do souboru pomoci
// metody write. Pokud vznikne vyjimka, zavola se automaticky destruktor 
// proudoveho objektu a ten soubor uzavre. 
// (Destruktor uzavre soubor i v pripade, ze vyjimka nevznikne a funkce
// Uloz2() skonci normalne.)

#include "seznam.h"
#include <stdio.h>
#include <fstream>

// Vyjmuti prvnich Kolik prvku ze seznamu a zapis do souboru
void Uloz(seznam<int> &S, int Kolik, const char* Soubor) throw (chyba_seznamu)
{
	FILE *f;
	f = fopen(Soubor, "ab");	// Otevri soubor v binarnim rezimu
	if(!f) return;				// Kdyz se to nepovede, konec
	try {						
		for(int i = 0; i < Kolik; i++)
		{						// Zkus vyjmout prvek a zapsat ho do souboru
			int j = S.VyjmiPrvni();
			fwrite(&j, sizeof(j), 1, f);
		}
	}
	catch(chyba_seznamu)		// a kdyz se to nepovede
	{
		fclose(f);				// uzavri soubor 
		cout << "Neco se deje" << endl;
		throw;					// a posli vyjimku dal.
	}
}

// Elegantnejsi reseni pomoci objektovych proudu
void Uloz2(seznam<int> &S, int Kolik, const char* Soubor) throw (chyba_seznamu)
{
	ofstream f(Soubor, ios::app|ios::binary);	// Vytvor proud pro pripisovani na konec, v binarnim rezimu
	if(!f) return;					// Kdyz se to nepovede, konec
	for(int i = 0; i < Kolik; i++)
	{
		int j = S.VyjmiPrvni();					// Vyjmi prvek a zapis ho
		f.write((const char*)&j, sizeof(j));
	}											// Zde se o vyjimky nestarame. Kdyz se to
}					// nepovede, destruktor soubor uzavre.


void main()
{
	seznam<int> S;
	for(int i = 0; i < 5; i++)
		S.VlozNaPocatek(i);
	try {
		Uloz(S, 10, "data.dta");
	}
	for(int i = 0; i < 5; i++)
		S.VlozNaPocatek(i);
	try {
		Uloz2(S, 10, "data.dta");
	}
	catch(exception& e)
	{
		cout << "Vyjimka: " << e.what() << endl;
	}
}