// Soubor SEZNAM.H - sablona jednosmerneho seznamu se zarazkou
//
// Obsahuje metody pro vkladani na pocatek a na konec seznamu,
// vyprazdneni seznamu, zruseni seznamu, pristup k hodnote ulozene v prvnim prvku,
// vyjmuti prvniho prvku seznamu a zjisteni, zda je seznam prazdny.
//
// V pripade neuspechu vyvolavaji nektere metody vyjimky typu chyba_seznamu:
// Jde o tyto metody: konstruktor, VlozNaPocatek a VlozNaKonec, pokud neni pamet
//     pro alokovany prvek
// VyjmiPrvni a obsahPrvniho, pokud je seznam prazdny
//
// Seznam nelze kopirovat. Toho lze dosahnout nasledujicim trikem:
// V tele tridy seznam deklarujeme kopirovaci konstruktor a prirazovaci
// operator jako soukrome metody. Jejich tela nedefinujeme, nejsou treba.
// Prekladac oznaci pokus o kopirovani jako chybu, protoze tyto metody nejsou
// dostupne.
// Na pocatku souboru muze byt #definovano makro LADIM. Je-li definovano, stane se soucasti
// tohoto souboru funkce main, ve ktere lze seznam testovat.

#define LADIM		
#include <iostream>
#include <stdexcept>

using namespace std;

class chyba_seznamu: public out_of_range {
public:
	chyba_seznamu(const string &s): out_of_range(s) {}
};

template<typename T> class prvek {
	T _hodnota;
	prvek * _dalsi;
public:
	prvek(T hodnota=0): _hodnota(hodnota), _dalsi(0) {}
	T &hodnota() { return _hodnota; }
	prvek *& dalsi() { return _dalsi; }
};

template<typename T> class seznam {
	prvek<T> *hlava, *zarazka;
	seznam(seznam&);			// Zakazujeme kopirovani
	seznam& operator=(seznam&);
public:
	seznam() throw(chyba_seznamu);	// Konstruktor vytvori prazdny seznam
	~seznam() throw();					// Destruktor vyprazdni seznam a zrusi i zarazku
	void Vyprazdni() throw();			// Odstrani ze seznamu vsechny prvky krome zarazky
	void VlozNaPocatek(T data) throw(chyba_seznamu);
	void VlozNaKonec(T data) throw(chyba_seznamu);
	T& ObsahPrvniho()  throw(chyba_seznamu);
	T VyjmiPrvni()  throw(chyba_seznamu);
	bool Prazdny() throw() { return hlava == zarazka; }
        prvek<T> * Najdi(T d) throw();
};

template<typename T>
seznam<T>::seznam() throw(chyba_seznamu): hlava(0), zarazka(0){	// Neskodna inicializace
        try {
        	hlava = new prvek<T>;						// Alokace zarazky
	        zarazka = hlava;						// V prazdnem seznamu ukazuje hlava i zarazka na tyz prvek - zarazku
        }
        catch(bad_alloc)
        {
                throw chyba_seznamu("nepodarilo se vytvorit seznam");
        }
}


template<typename T>
seznam<T>::~seznam() throw() {
	Vyprazdni();			// Vyprazdni seznam
	delete hlava;			// Odstran zarazku
}

template<typename T>		// Odstrani vsechny prvky krome zarazky
void seznam<T>::Vyprazdni() throw(){
	while(hlava != zarazka)
	{
		prvek<T>* p = hlava;	// Uloz ukazatel na prvni prvek
		hlava = p->dalsi();	// Do ukazatele na prvni prvek uloz adresu druheho
		delete p;		// Vymaz prvek, ktery byl prvni
	}
}

template<typename T>
void seznam<T>::VlozNaPocatek(T d) throw(chyba_seznamu) {
	try{ 
		prvek<T> *p = new prvek<T>(d);	// Alokuj novy prvek
		p->dalsi() = hlava;		// Jeho ukazatel na nasledujici bude ukazovat na prvni prvek seznamu	
		hlava = p;			// Ukazatel na prvni prvek seznamu ukazuje na novy prvek
	}
	catch(bad_alloc)
	{
		throw chyba_seznamu("neni pamet pro novy prvek");
	}
}

template<typename T>
T& seznam<T>::ObsahPrvniho() throw (chyba_seznamu)
{
	if(Prazdny())
		throw chyba_seznamu("nelze menit prvni prvek prazdneho seznamu");
	else
		return hlava->hodnota();
}

template<typename T>
void seznam<T>::VlozNaKonec(T d) throw(chyba_seznamu)
{
	try {
		prvek<T>* p = new prvek<T>(0);	// Alokuj novy prvek
		zarazka->hodnota() = d;			// Uloz data do zarazky
		zarazka->dalsi() = p;			// Novy prvek pripoj jako zarazku
		zarazka = p;
	}
	catch(bad_alloc)
	{
		throw chyba_seznamu("neni pamet ");
	}
}

template<typename T>
T seznam<T>::VyjmiPrvni() throw (chyba_seznamu)
{
	if(Prazdny()) throw chyba_seznamu("Vyber prvniho prvku z prazdneho seznamu");
	prvek<T> *p = hlava;	// Zapamatuj si ukazatel na prvni prvek seznamu
	T d = p -> hodnota();	// Zapamatuj si data z prvniho prvku
	hlava = p->dalsi();	// Ukazatel na prvni prvek bude ukazovat na druhy
	delete p;		// Vymaz byvaly prvni prvek seznamu
	return d;		// Vrat zapamatovanou hodnotu
}

template<typename T>
prvek<T> * seznam<T>::Najdi(T d) throw()
{
   zarazka -> hodnota() = d;     // Uloz hledanou hodnotu do zarazky (takze je jiste, ze v seznamu je)
   prvek<T>* p = hlava;          // Pomocny ukazatel na aktualni prvek seznamu - zaciname u prvniho prvku
   while(p->hodnota() != d)      // Pokud p neukazuje na prvek obsahujici d, jdi na dalsi
      p = p->dalsi();
   return p == zarazka ? 0 : p;  // Ukazuje-li p na zarazku, hledana hodnota v seznamu neni, vrat NULL
}                                // Jinak vrat adresu prvku

#ifdef LADIM
// Pomocna funkce main pro ladeni seznamu
void main(){
	seznam<int> S;
    int i;
	for(i = 1; i < 5; i++)
		S.VlozNaPocatek(i);
	S.ObsahPrvniho() = 10;
	S.VlozNaKonec(12);
	i = S.VyjmiPrvni();
    prvek<int> *p = S.Najdi(3);
	S.Vyprazdni();
        // seznam<int> T(S); // Chyba - kopirovaci konstruktor je nepristupny
        seznam<int> U;
        //U = S;             // Chyba - prirazovaci operator je nepristupny
}

#endif