// Soubor DCAST1.CPP
// Priklady ukazujici fungovani operatoru dynamic_cast
//
// Nejprve tridy, na kterych to budeme predvadet - musi byt polymorfni,
// tak jim dame virtualni destruktor; metoda hodnota neni virtualni

#include <typeinfo>
#include <iostream>
using namespace std;

class A {
	int a;
public:
	A(int i): a(i){}
	virtual ~A() {}
	int hodnota() { return a; }
};

class B {
	int b;
public:
	B(int i): b(i){}
	virtual ~B() {}
	int hodnota() { return b; }
};

// C je spolecnym (jednoznacnym a pristupnym) potomkem A a B

class C: public A, public B
{
	int c;
public: 
	C(int c): A(c), B(c+1), c(c+2) {}
	int hodnota() { return c; }
};

void main()
{
	A a(1); // Instance
	B b(2);
	C c(3);
	A* ua;	// Ukazatele na instance
	B* ub;
	C* uc = &c;
	
// Nejprve priklady, kde je pouziti operatoru dynamic_cast dovolene, ale zbytecne
//
	ua = dynamic_cast<A*>(uc);	// Pretypovani na predka: dovolene, i kdyz zbytecne
	ua = uc;					// Znamena naprosto totez co predchozi prikaz 
							// nebo ua = (A*)uc;
							// nebo ua = static_cast<A*>(uc);

	int i = dynamic_cast<A&>(c).hodnota();		// Zpusobi totez co nasledujici prikazy
	i = ((A&)c).hodnota();				// Totez co predchozi prikaz
	i = static_cast<A&>(c).hodnota();		// Totez
	i = c.A::hodnota();				// Totez bez pretypovani

// Trochu uzitecnejsi pretypovani: Konverze na void*
	ub = uc;					// ub ukazuje na zdedeny podobjekt typu B v c
	void *v1 = dynamic_cast<void*>(ub);		// v1 obsahuje adresu c
	void *v2 = static_cast<void*>(ub);		// v2 obsahuje adresu zdedeneho podpobjektu typu B, v1 != v2
	cout << ((v1 == v2) ? "Ano" : "Ne") << endl;	// Vypise Ne

// Pretypovani na potomka
	ub = uc;					// ub ukazuje na zdedeny podobjekt typu B v c

	C* ucc = dynamic_cast<C*>(ub);			// Pretypovani se podari, nebot ub obsahuje adresu zdedeneho podobjektu v instanci tridy C
	if(ucc)						// Je treba kontrolovat, zda se to podarilo
		cout << ucc -> hodnota() << endl;
	else
		cout << "nepovedlo se" << endl;

	ub = &b;					// ub ukazuje na samostatnou instanci

	ucc = dynamic_cast<C*>(ub);			// Pretypovani se nepodari, nebot ub obsahuje adresu samostatne instance tridy B
	if(ucc)						// Je treba kontrolovat, zda se to podarilo
		cout << ucc -> hodnota() << endl;	// Vypise 5
	else
		cout << "nepovedlo se" << endl;		// a tentokrat se dozvime, ze ne

// Totez pro reference
	B &rb = c;					// rb je odkaz na zdedeny podobjekt tridy B v instanci c
	cout << rb.hodnota() << endl;			// Vypise 4
	
	try {
		cout << dynamic_cast<C&>(rb).hodnota() << endl;	// Vypise 5, nebot rb je reference na zdedeny podobjekt v instanci typu C
		cout << dynamic_cast<C&>(b).hodnota() << endl;	// Nepovede se, nebot b je samostatna instance - vznikne vyjimka
	}
	catch(...)
	{
		cout << "Pretypovani se nepovedlo" << endl;
	}

// Pretypovani na z predka na jineho nezavisleho predka

	ub = uc;					// ub ukazuje na zdedeny podobjekt typu B v c
	ua = dynamic_cast<A*>(ub);			// Lze - ub obsahuje adresu zdedeneho podobjektu tridy, ktera ma za predka take A
	if(ua)
		cout << ua->hodnota() << endl;		// Vypise 3  (Je rozumne opet zjistovat, zda se pretypovani podarilo.)

	ua = &a;					// Nyni obsahuje ua adresu samostatne instance
	ub = dynamic_cast<B*>(ua);			// takze toto pretypovani se nepovede
	if(ub)
		cout << ub->hodnota() << endl;		
	else
		cout<< "Pretypovani se nepovedlo" << endl;

	ub = (B*)(C*)ua;				// Klasicka cesta
	cout << ub->hodnota() << endl;			// Vypise nejaky nesmysl, nebot ua ukozovalo na samostatnou instanci a pretupovani na B* nemelo smysl
}
