// Soubor STL03.CPP
// Priklad pouziti alokatoru
/* Alokatory jsou objekty, ktere ridi alokaci pameti pro kontejnery.
   Je to instance sablony, ktera musi mit v podstate nasledujici tvar.

   Alokaci obstarava metoda allocate, uvolnovani pameti
   metoda deallocate. Protoze zde nam jde o alokaci do
   "areny", pole znaku, je metoda deallocate prazdna.

   Program vytvori instanci D tridy deque s alokatorem pouzivajicim pole C. Do teto
   dvoustranne fronty ulozi znaky male a velke abecedy a pak toto pole vypise do souboru.

   Ve vypisu pole do souboru se lze presvedsit, ze D skutecne alokuje sve prvky v C.

   Pozor, tento priklad se lisi pro prekladace, ktere podporuji
   vnorene sablony (C++Builder 4 nebo novejsi) a v prekladacich, ktere je nepodporuji.

   V MS Visiual C++ 6 je vnorena sablona nahrazena funkci _Charalloc().

   Tento program obsahuje verzi jak pro MSVC tak i pro C++Builder. Jsou rozliseny pomoci
   direktiv #ifdef:
   __BCPLUSPLUS__ je pro C++Builder
   _MSC_VER pro VS Visual C++
   (Uvedeny test neni dostatecny - meli bychom jeste testovat verzi prekladace, napr. 
    _MSC_VER = 1200 (MSVC 6.0) - zde neni jasne, jak bude vypadat pristi verze, takze rovnost
    nebo __BCPLUSPLUS__ >= 0x0540 zde implementace odpovida standardu, takze lze pouzit >=.
   To prenechavam ctenarum.)
*/

#include <deque>
#include <fstream>
const int N = 10000;
using namespace std;

char C[N];

#ifdef __BCPLUSPLUS__

// Alokator pro Borland C++ Builder (vlastne podle standardu)
template <class T> class alloc {	
  int index;	// Index urcujici zacatek pristi alokace **
public:
  typedef size_t            size_type;
  typedef ptrdiff_t         difference_type;
  typedef T*                pointer;
  typedef const T*          const_pointer;
  typedef T&                reference;
  typedef const T&          const_reference;
  typedef T                 value_type;
  template <class U> struct rebind {
    typedef alloc<U> other;
  };
  alloc () throw():index(0){};
  //alloc (const alloc&) throw ();

  template <class U> alloc(const alloc<U>& A) throw():index(0) {};
  template <class U>   // **
    alloc& operator=(const alloc<U>&) throw();
  ~alloc () throw(){};
  pointer address (reference x) const {return &x;};
  const_pointer address (const_reference) const {return &x;};
  pointer allocate (size_type n, void* = 0)
  {
        void *p = &C[index];
        index+= n;
        return (pointer)p;
  }
  void deallocate(pointer p, size_type n){};
  size_type max_size () const {return N;};
  void construct (pointer p, const T& t){*p = t;};
  void destroy (pointer) {}
};

#elif defined _MSC_VER

// Alokator pro MS Visual C++ (s funkci _Charalloc())
template<class _Ty>
	class alloc {
	int index;
public:
	alloc():index(0){}
	typedef unsigned size_type;
	typedef int difference_type;
	typedef _Ty *pointer;
	typedef const _Ty *const_pointer;
	typedef _Ty & reference;
	typedef const _Ty & const_reference;
	typedef _Ty value_type;
	pointer address(reference _X) const
		{return (&_X); }
	const_pointer address(const_reference _X) const
		{return (&_X); }
	pointer allocate(size_type n, const void *)
	{
		void *p = &C[index];
	    index+= n;
		return (pointer)p;
	}
	char *_Charalloc(size_type n)
	{
	    void *p = &C[index];
        index+= n;
        return (pointer)p;
	}
	void deallocate(void *_P, size_type)
		{ }
	void construct(pointer p, const _Ty& t)
		{*p = t; }
	void destroy(pointer _P)
		{ }
	size_type max_size() const
		{size_type _N = (size_type)(-1) / sizeof (_Ty);
		return (0 < _N ? _N : 1); }
	};

#endif

void main()
{
  deque<char, alloc<char> > D;
  for(int i = 'A'; i < 'z'; i++) D.push_front(char(i));
  char cc = D[1];
  ofstream F("data.dta");
  F.write(C, N);
  F.close();
}